from itertools import combinations
from collections import defaultdict
import numpy as np
import copy

class Graph: 
    """
    get graph from a list of edges.
    note that if a fully-connected graph is NOT a DAG,
    then the smallest graph (a triangle) in it must also be a DCG.
    """
    def __init__(self, vertices): 
        #self.graph = defaultdict(list) 
        self.graph = {} 
        self.start_nodes = set() 
        #-------- for equal chain --------#
        self.eq_map_in = {}
        self.eq_map_out = {}
        #--------------------------------#        
        #-------- for SCC --------#
        self.V = vertices
        self.Time = 0
        #-------------------------#
  
    def addEdge_eq_as_new_nodes(self,u, v, equal=False):
        if equal:
            if u not in self.graph:
                self.graph.update({u: {'in': set(), 'out': set(), 'eq': set()}})
            if v not in self.graph:
                self.graph.update({v: {'in': set(), 'out':set(), 'eq': set()}})
            self.graph[u]['eq'].add(v)
            self.graph[v]['eq'].add(u)
        else:
            if u not in self.graph:
                self.graph.update({u: {'in': set(), 'out': set(), 'eq': set()}})
                self.start_nodes.add(u)
            if v not in self.graph:
                self.graph.update({v: {'in': set(), 'out':set(), 'eq': set()}})
            self.graph[u]['out'].add(v)
            self.graph[v]['in'].add(u)
            self.start_nodes.discard(v)

    def get_adj(self):
        pass

    def addEdge(self,u, v, equal=False, value=None):
        u = self.eq_map_out[u] if u in self.eq_map_out else u
        v = self.eq_map_out[v] if v in self.eq_map_out else v
        if u == v: return
        if equal:
            # things must to know about equal:
            # 1. equal is NOT transitive, this is caused by the unreliable threshold.
            # 2. equal is symmetric.
            # 3. for a fully-connected graph, there is at most one edge between two edges,
            #    which results in the egde-graph a 2-hop DAG.
            #    proof: for any two nonadjacent edges AB CD, we can connect AC BD, or AD BC two make them connected.
            if u not in self.graph:
                self.graph.update({u: {'in': set(), 'out': set(), 'eq': set()}})
                self.start_nodes.add(u)
            if v not in self.graph:
                self.graph.update({v: {'in': set(), 'out':set(), 'eq': set()}})
                self.start_nodes.add(v)
            # self.graph[u]['eq'].add(v)
            # self.graph[v]['eq'].add(u)

            #------------- eq as bi-directed edge ----------------#
            self.graph[u]['in'].add(v) 
            self.graph[u]['out'].add(v)
            self.graph[v]['in'].add(u)
            self.graph[v]['out'].add(u)

            if value < 0:
                self.graph[u]['eq'].add((v, -1*value))
            else:
                self.graph[v]['eq'].add((u, value))
            #-----------------------------------------------------#
            #------------- eq with ['in'] pointing to ----------------#
            #self.addEqualWithInPointingTo_eq(u, v)
            #---------------------------------------------------------#
            #-------------- add equal nodes with an equal chain ---------------#
            #self.addEqualWithEqualChain(u, v)
            #------------------------------------------------------------------#

        else:
            if u not in self.graph:
                self.graph.update({u: {'in': set(), 'out': set(), 'eq': set()}})
                self.start_nodes.add(u)
            if v not in self.graph:
                self.graph.update({v: {'in': set(), 'out':set(), 'eq': set()}})
            self.graph[u]['out'].add(v)
            self.graph[v]['in'].add(u)
            self.start_nodes.discard(v)
        

            #------------- eq with ['in'] pointing to ----------------#
            #self.addEqualWithInPointingTo_in(u, v)
            #---------------------------------------------------------#
            #------------- eq ----------------#
            #----------------------------------#

    def addEqualWithInPointingTo_eq(self, u, v):
        self.graph[u]['in'].discard(v)
        self.graph[u]['out'].discard(v)
        self.graph[v]['in'].discard(u)
        self.graph[v]['out'].discard(u)

        for u_in in self.graph[u]['in']:
            if u_in not in self.graph[v]['eq'] and u_in not in self.graph[v]['out']:
                self.graph[u_in]['out'].add(v)
                self.graph[v]['in'].add(u_in)
                self.start_nodes.discard(v)
        for v_in in self.graph[v]['in']:
            if v_in not in self.graph[u]['eq'] and v_in not in self.graph[u]['out']:
                self.graph[v_in]['out'].add(u)
                self.graph[u]['in'].add(v_in)
                self.start_nodes.discard(u)

    def addEqualWithInPointingTo_in(self, u, v):
        self.graph[u]['in'].discard(v)
        self.graph[v]['out'].discard(u)
        for v_eq in self.graph[v]['eq']:
            if v_eq not in self.graph[u]['eq']:
                self.graph[u]['out'].add(v_eq)
                self.graph[v_eq]['in'].add(u)
                self.start_nodes.discard(v_eq)

    def addEqualWithEqualChain(self, u, v):
        # max node -> min node
        min_node = min(u, v)
        max_node = max(u, v)
        if min_node not in self.graph:
            if len(self.graph) == 0:
                self.start_nodes.add(min_node)
            self.graph.update({min_node: {'in': set(), 'out': set(), 'eq': set()}})
        self.graph[min_node]['eq'].add(max_node)
        # replace the start node
        if max_node in self.start_nodes:
            self.start_nodes.remove(max_node)
            self.start_nodes.add(min_node)
        # the in-out edges and eq nodes of max node are added to min node
        if max_node in self.graph:
            self.graph[min_node]['in'] |= self.graph[max_node]['in']
            self.graph[min_node]['out'] |= self.graph[max_node]['out']
            self.graph[min_node]['eq'] |= self.graph[max_node]['eq']
            # can be deleted if eq node inserted first
            for node in self.graph[max_node]['in']:
                self.graph[node]['out'].add(min_node)
                self.graph[node]['out'].discard(max_node)
            for node in self.graph[max_node]['out']:
                self.graph[node]['in'].add(min_node)
                self.graph[node]['in'].discard(max_node)

            del self.graph[max_node]
        # max node -> min node
        if min_node not in self.eq_map_in:
            self.eq_map_in.update({min_node: set()})
        self.eq_map_in[min_node].add(max_node)
        self.eq_map_out.update({max_node: min_node})
        # nodes to max node -> nodes to min node
        if max_node in self.eq_map_in:
            self.eq_map_in[min_node] |= self.eq_map_in[max_node]
            for node in self.eq_map_in[max_node]:
                self.eq_map_out[node] = min_node
            del self.eq_map_in[max_node]
    
    def copy(self, vertices=None):
        if vertices is None:
            return copy.deepcopy(self)
        else:
            new_graph = Graph(max(vertices) + 1) 
            for v in vertices:
                new_graph.graph[v] = copy.deepcopy(self.graph[v]) 
            return new_graph
    
    def isReachable(self, s, d):
        if s not in self.graph or d not in self.graph:
            return False
        # Mark all the vertices as not visited
        visited =[False]*(self.V)
  
        # Create a queue for BFS
        queue=[]
  
        # Mark the source node as visited and enqueue it
        queue.append(s)
        visited[s] = True
  
        while queue:
 
            #Dequeue a vertex from queue
            n = queue.pop(0)
             
            # If this adjacent node is the destination node,
            # then return true
            if n == d:
                   return True
 
            #  Else, continue to do BFS
            for i in self.graph[n]['out']:
                if i in self.graph and visited[i] == False:
                    queue.append(i)
                    visited[i] = True
         # If BFS is complete without visited d
        return False

    def isReachableWithoutEq(self, s, d, eq_edges):
        path = [s]

        # Create a queue for BFS
        queue = []
  
        # Mark the source node as visited and enqueue it
        queue.append(path)

        self.visited_edges = 0
  
        while queue:
 
            #Dequeue a vertex from queue
            tmp_path = queue.pop(0)
            n = tmp_path[-1]
             
            # If this adjacent node is the destination node,
            # then return true
            if n == d:
                # print(len(tmp_path))
                # print(tmp_path)
                if not self.onlyEqEdges(tmp_path, eq_edges):
                    return True
 
            #  Else, continue to do BFS
            for i in self.graph[n]['out']:
                self.visited_edges += 1
                if i in self.graph and i not in tmp_path:
                    queue.append(tmp_path + [i])
         # If BFS is complete without visited d
        return False
    
    def haveEdge(self, n1, n2):
        if n1 not in self.graph or n2 not in self.graph:
            return False
        if n2 in self.graph[n1]['in'] or n2 in self.graph[n1]['out']:
            return True
        else:
            return False
            

    def onlyEqEdges(self, path, eq_edges):
        edges = list(combinations(path, 2))
        for n1, n2 in edges:
            if self.haveEdge(n1, n2) and (n1, n2) not in eq_edges:
                return False
        return True


        
    def SCCUtil(self, u, low, disc, stackMember, st, scc, vertices=None, size=2):
        # Initialize discovery time and low value
        disc[u] = self.Time
        low[u] = self.Time
        self.Time += 1
        stackMember[u] = True
        st.append(u)
 
        # Go through all vertices adjacent to this
        for v in self.graph[u]['out']:
 
            if vertices is not None:
                if v not in vertices:
                    continue
            # If v is not visited yet, then recur for it
            if disc[v] == -1:
 
                self.SCCUtil(v, low, disc, stackMember, st, scc, vertices, size)
 
                # Check if the subtree rooted with v has a connection to
                # one of the ancestors of u
                # Case 1 (per above discussion on Disc and Low value)
                low[u] = min(low[u], low[v])
 
            elif stackMember[v] == True:
 
                '''Update low value of 'u' only if 'v' is still in stack
                (i.e. it's a back edge, not cross edge).
                Case 2 (per above discussion on Disc and Low value) '''
                low[u] = min(low[u], disc[v])
 
        # head node found, pop the stack and print an SCC
        w = -1  # To store stack extracted vertices
        if low[u] == disc[u]:
            tmp = []
            while w != u:
                w = st.pop()
                # print(w, end=" ")
                tmp.append(w)
                stackMember[w] = False
            if len(tmp) > size:
                scc.append(tmp)
            # print()
 
    # The function to do DFS traversal.
    # It uses recursive SCCUtil()
 
    def SCC(self, vertices=None, size=2):
        # Mark all the vertices as not visited
        # and Initialize parent and visited,
        # and ap(articulation point) arrays
        self.Time = 0
        disc = [-1] * (self.V)
        low = [-1] * (self.V)
        stackMember = [False] * (self.V)
        st = []
        scc = []
        # Call the recursive helper function
        # to find articulation points
        # in DFS tree rooted with vertex 'i'
        if vertices is None:
            for i in range(self.V):
                if i not in self.graph:
                    continue
                if disc[i] == -1:
                    self.SCCUtil(i, low, disc, stackMember, st, scc, size=size)
        else:
            for i in vertices:
                if disc[i] == -1:
                    self.SCCUtil(i, low, disc, stackMember, st, scc, vertices, size=size)
        return scc

#    def topologicalSortUtil(self,v,visited,stack): 
#  
#        visited[v] = True
#  
#        for i in self.graph[v]: 
#            if visited[i] == False: 
#                self.topologicalSortUtil(i,visited,stack) 
#  
#        stack.insert(0,v) 
#  
#    def topologicalSort(self): 
#        visited = [False]*self.V 
#        stack =[] 
#  
#        for i in range(self.V): 
#            if visited[i] == False: 
#                self.topologicalSortUtil(i,visited,stack) 


class UnionFindSet():
    def __init__(self, data_list):
        self.father_dict = {}
        self.size_dict = {}
        self.different_sets = {}
        self.data_list = data_list
        nodes = np.array(data_list).flatten().tolist()
        for node in nodes:
            self.father_dict[node] = node
            self.size_dict[node] = 1
            self.different_sets.update({node: set([node])})
 
    def find(self, node):
        father = self.father_dict[node]
        if(node != father):
            if father != self.father_dict[father]:
                self.size_dict[father] -= 1
            father = self.find(father)
        self.father_dict[node] = father
        return father
 
    def is_same_set(self, node_a, node_b):
        return self.find(node_a) == self.find(node_b)
 
    def union(self, node_a, node_b):
        if node_a is None or node_b is None:
            return
        a_root = self.find(node_a)
        b_root = self.find(node_b)
 
        if(a_root != b_root):
            a_set_size = self.size_dict[a_root]
            b_set_size = self.size_dict[b_root]
            if(a_set_size >= b_set_size):
                self.father_dict[b_root] = a_root
                self.size_dict[a_root] = a_set_size + b_set_size
                self.different_sets[a_root] |= self.different_sets[b_root]
                del self.different_sets[b_root]
            else:
                self.father_dict[a_root] = b_root
                self.size_dict[b_root] = a_set_size + b_set_size
                self.different_sets[b_root] |= self.different_sets[a_root]
                del self.different_sets[a_root]
        
    def init_union(self):
        for i in range(len(self.data_list)):
            node1, node2 = self.data_list[i][0], self.data_list[i][1]
            self.union(node1, node2)


if __name__ == '__main__':
    g = Graph(6)
    g.addEdge(0, 1)
    g.addEdge(2, 3)
    g.addEdge(2, 4)
    g.addEdge(3, 4)
    g.addEdge(0, 5)
    print(g.graph)
    print(g.isReachableWithoutEq(2,4,[]))