from treebuilder import TreeBuilder
from itertools import combinations
from utils import read_data, equal_vertices_judgement, simple_cycles
import sys

# integer programming
import networkx as nx
from cvxopt import matrix
from cvxopt.glpk import ilp
import numpy as np
import os

DEBUG = False 
args = {
    'change_greenstar': True,
    'threshold1': 3, # red star threshold
    'threshold2': 0, # gene flow display threshold
    'threshold3': 3, # gene flow merge threshold 
    'bidirectional': True, # calculate the reversed gene flow
    'calculate_pd': True, # calculate the gene flow confidence
    'data_path': 'data/D_AI_230710.log',
    'save_path': 'pictures/tree.png',
    'draw_gene_flow': True,
    'gene_flow_display_confidence': 0.7,
    'target_block_list': ['AR14K'],
    's1_block_list': ['AR14K'],
    # 'target_block_list': [],
    # 's1_block_list': [],
}
populations = ['Tianyuan',
    'Boshan',
    'Qihe',
    'Longlin',
    'Dushan',
    'Baojianshan',
    'La368',
    'Bianbian',
    'Xiaojingshan',
    'BQGroup',
    'YJCGroup',
    'LJZGroup',
    'YR_MN_WGM',
    'WLR_MN',
    'DevilsCave_N',
    'AR14K']

# populations = ['Kostenki14',
#                'Tianyuan',
#                'AR19K',
#                'AR14K',
#                'Zongri5_1k',
#                'Boshan',
#                'Qihe3',
#                'Donghulin',
#                ]

#args = {
#    'change_greenstar': True,
#    'threshold1': 3, # red star threshold
#    'threshold2': 0, # gene flow display threshold
#    'threshold3': 3, # gene flow merge threshold 
#    'bidirectional': True, # calculate the reversed gene flow
#    'calculate_pd': False, # calculate the gene flow confidence
#    'data_path': 'data/sub.f4.GX.log',
#    'save_path': 'pictures/tree.png',
#    'draw_gene_flow': True,
#}
#
#populations = ['Kostenki14',
#               'Tianyuan',
#               'Boshan',
#               'Qihe',
#               'Longlin',
#               'Dushan',
#               'Baojianshan',
#               'La368'
#               ]


#data_path = 'data/f4_20230111.log'
#populations = [
#    'Kostenki14',
#    'Tianyuan',
#    'Yana',
#    'Liangdao2',
#    'Qihe',
#    'Yumin',
#    'Boshan',
#    'Xiaogao',
#    'Kolyma',
#    'Shamanka_EN',
#    'Longlin',
#    'Dushan',
#    'Baojianshan',
#    'La368'
#]
#
#data_path = 'data/f4.YN.log'
#populations = [
#    'Mbuti',
#    'Tianyuan',
#    'La368',
#    'Longlin',
#    'Xingyi_EN',
#    'Xingyi_LN',
#    'Qihe3',
#    'Boshan',
#    'Baiyangcun'
#]

change_greenstar = args.get('change_greenstar', True) 
threshold1 = args.get('threshold1', 3)
threshold2 = args.get('threshold2', 0)
threshold3 = args.get('threshold3', 3)
data_path = args.get('data_path')
save_path = args.get('save_path')
print_gene_flow = args.get('draw_gene_flow', True)

threshold = threshold1
gene_flow_threshold = threshold3


os.makedirs('/'.join(save_path.split('/')[:-1]), exist_ok=True)
relations = list(combinations(sorted(populations), 2))

#--------------- check scc with threshold = 0 ------------------#
print()
print('Analyzing the minimum Z...')
# get all edges
bottom_threshold = 0
D_statistics, adjusted_D_values, triangles, redstar_indices, min_redstar, s2_record \
    = read_data(data_path, bottom_threshold, populations)
print(s2_record)
builder = TreeBuilder(populations, triangles, s2_record)
builder.D2graph(D_statistics)
g = builder.g
scc0 = g.SCC()
num_scc0 = len(scc0)
if DEBUG:
    print('num_scc0', num_scc0)
    print('size of s0:', [len(s) for s in scc0])

unchangeable_edges = []
for s in scc0:
    for idx in s:
        # the inversed edge is unchangeable
        for node_idx in g.graph[idx]['in']:
            if node_idx in s:
                unchangeable_edges.append((idx, node_idx))
for e in unchangeable_edges:
    e1 = set(builder.relations[e[0]])
    e2 = set(builder.relations[e[1]])
    common_pop = (e1 & e2).pop()
    key = tuple(sorted(list(e1 | e2)))
    value = triangles[key]['D'][key.index(common_pop)]
    bottom_threshold = max(bottom_threshold, abs(value))
    if key.index(common_pop) != redstar_indices[key]:
        print('Warning: changing the value which is not the red star!')
        if not change_greenstar:
            raise ValueError('Error: this graph cannot be plotted without changing the green stars, please set green=True!')
        
        for i in range(key.index(common_pop), key.index(common_pop)+3):
            print(key[(i+1)%3], key[(i+2)%3], key[i%3], triangles[key]['D'][i%3])
if DEBUG:
    print('bottom_threshold', bottom_threshold)
    print('unchangeable_edges', len(unchangeable_edges))
print()
print('Minimum Z:', bottom_threshold)

if threshold < bottom_threshold:
    ValueError('the given Z is smaller than the minimum Z!')

#---------------------------------------------------------------#
D_statistics, adjusted_D_values, triangles, redstar_indices, min_redstar, s2_record \
    = read_data(data_path, threshold, populations)
if DEBUG:
    if min_redstar:
        print('min red star:', min(min_redstar))
    print(adjusted_D_values)
builder = TreeBuilder(populations, triangles, s2_record)
builder.D2graph(D_statistics)
g = builder.g


print()
print('Adjusting values...')

eq_scc0 = {}
for scc_idx, s in enumerate(scc0):
    # if len(s) <= 2: continue
    if DEBUG:
        print('-------------- scc ----------------')
        print(s)
    represent = s[-1]
    for idx in s[:-1]:
        for node in g.graph[idx]['in']:
            if node in s: continue
            g.graph[represent]['in'].add(node)
            g.graph[node]['out'].add(represent)
            g.graph[node]['out'].remove(idx)
            # note that, the eq set of the node may contain duplicated nodes with different values!!!!!!!
            # tmp_node_record is used to duplicate them.
#            tmp_node_record = []
            new_eq_set = set()
            for eq_node, value in g.graph[node]['eq']:
#                if eq_node in tmp_node_record: continue
                if eq_node == idx:
                    eq_node = represent
                new_eq_set.add((eq_node, value))
#                tmp_node_record.append(eq_node) 
            g.graph[node]['eq'] = new_eq_set
        for node in g.graph[idx]['out']:
            if node in s: continue
            g.graph[represent]['out'].add(node)
            g.graph[node]['in'].add(represent)
            g.graph[node]['in'].remove(idx)
        for node, value in g.graph[idx]['eq']:
            if node in s: continue
            g.graph[represent]['eq'].add((node, value))
        del g.graph[idx]
        g.graph[represent]['in'].discard(idx)
        g.graph[represent]['out'].discard(idx)
        if idx in g.start_nodes:
            g.start_nodes.discard(idx)
        else:
            g.start_nodes.discard(represent)
    # note that, the eq set of the node may contain duplicated nodes with different values!!!!!!!
    # tmp_node_record is used to duplicate them.
#    tmp_node_record = []
    new_eq_set = set()
    for node, value in g.graph[represent]['eq']:
        if node in s: continue #or node in tmp_node_record: continue
        new_eq_set.add((node, value))
#        tmp_node_record.append(node)
#    del tmp_node_record
    g.graph[represent]['eq'] = new_eq_set
    if DEBUG:
        print(g.graph[represent]['in'])
        print(g.graph[represent]['out'])
        print(g.graph[represent]['eq'])
    eq_scc0[represent] = s[:-1]

if DEBUG:
    print('start_nodes', g.start_nodes)
scc = g.SCC()
num_scc = len(scc)
if DEBUG:
    print('num_scc', num_scc)
    print('size of s:', [len(s) for s in scc])

num_adjusted = []
total_edges_adjusted = []
eq_edges_map = {}
for scc_idx, s in enumerate(scc):
    # if scc_idx != 1: continue
    print()
    print('Adjusting %d/%d scc...' % (scc_idx + 1, num_scc))
    if DEBUG:
        print(s)
    edges_to_adjust = []
    edges_to_adjust_with_value = []
    eq_edges = []
    eq_nodes = set()
    nodes_to_adjust = set()
    edges_to_adjust_weights = []
    for idx in s:
        tmp_in = set()
        tmp_out = set()
        tmp_eq = set() 
        for n in g.graph[idx]['in']:
            if n in s:
                tmp_in.add(n)
        for n in g.graph[idx]['out']:
            if n in s:
                tmp_out.add(n)
        for n, v in g.graph[idx]['eq']:
            if n in s:
                tmp_eq.add(n)
#        print('----------------')
#        print(tmp_out)
#        print(tmp_eq)
        if tmp_out != tmp_in:
                eq_nodes.add(idx)
        for node_idx, value in g.graph[idx]['eq']:
            if node_idx in s:
                eq_edges.append((idx, node_idx))
                eq_edges.append((node_idx, idx))
            if node_idx in s and (idx, node_idx) not in unchangeable_edges:
                # print('eq:', idx, node_idx)
                if (idx, node_idx) not in edges_to_adjust:
                    edges_to_adjust_weights.append(1)
                    edges_to_adjust.append((idx, node_idx))
                    nodes_to_adjust.add(idx)
                    nodes_to_adjust.add(node_idx)
                    edges_to_adjust_with_value.append((idx, node_idx, value))
                    eq_edges_map[(idx, node_idx)] = [(idx, node_idx)]
                else:
                    edges_to_adjust_weights[edges_to_adjust.index((idx, node_idx))] += 1
                    eq_edges_map[(idx, node_idx)].append((idx, node_idx))
    if DEBUG: 
        print('calculating adjusted...')
    if DEBUG:
        print(len(s), len(edges_to_adjust_with_value))
        print(len(eq_edges))
    adjusted = []
    for e_idx, (node_o, node_i, value) in enumerate(edges_to_adjust_with_value):
        g_tmp = g.copy(s)
        g_tmp.graph[node_o]['out'].discard(node_i)
        g_tmp.graph[node_i]['out'].discard(node_o)
        g_tmp.graph[node_i]['in'].discard(node_o)
        g_tmp.graph[node_o]['in'].discard(node_i)
        g_tmp.graph[node_o]['eq'].discard((node_i, value))
        for e_idx2, (node_o2, node_i2, value2) in enumerate(edges_to_adjust_with_value):
            if e_idx2 == e_idx: continue 
            g_tmp.graph[node_o2]['out'].discard(node_i2)
            g_tmp.graph[node_i2]['in'].discard(node_o2)
            g_tmp.graph[node_o2]['eq'].discard((node_i2, value2))
        if g_tmp.isReachableWithoutEq(node_i, node_o, eq_edges):
            adjusted.append((node_o, node_i, value))
    if DEBUG:
        print('adjusted', len(adjusted))
        print(adjusted)
    num_adjusted.append(len(adjusted))
    for e_idx, (node_o, node_i, value) in enumerate(adjusted):
        g.graph[node_o]['out'].remove(node_i)
        g.graph[node_i]['in'].remove(node_o)
        edges_to_adjust.remove((node_o, node_i))
        edges_to_adjust_with_value.remove((node_o, node_i, value))

        # count adjusted edges
        for e in eq_edges_map[(node_o, node_i)]:
            n1, n2 = e
            r1 = set(builder.relations[n1])
            r2 = set(builder.relations[n2])
            if not (r1 & r2): continue
            common_pop = (r1 & r2).pop()
            key = tuple(sorted(list(r1 | r2)))
            k_idx = key.index(common_pop)
            value = triangles[key]['D'][k_idx]
            total_edges_adjusted.append((key[(k_idx+1)%3], key[(k_idx+2)%3], key[k_idx], value))
    # print('adjusted', adjusted)
    if DEBUG:
        print(equal_vertices_judgement(g, s))
        print(g.SCC(s))
        print(s)

    G = nx.DiGraph()
    for node_idx in s:
        G.add_node(node_idx)
        for node_idx2 in g.graph[node_idx]['in']:
            if node_idx2 in s:
                G.add_edge(node_idx2, node_idx)
    if DEBUG:
        print('here')
        print('num of nodes:', len(G.nodes))
        print('num of edges:', len(G.edges))
        print('num of adjust nodes:', len(nodes_to_adjust))
        print('num of adjust edges:', len(edges_to_adjust))
        print(edges_to_adjust)
        # print('eq_edges', eq_edges)
        print('eq_edges', len(eq_edges))
        print('eq_nodes', len(eq_nodes))
    cycles = list(simple_cycles(G, eq_edges, edges_to_adjust))

    num_edges = len(edges_to_adjust)
    num_cycles = len(cycles)
    if DEBUG:
        print(num_edges)
        print(num_cycles)
        print(edges_to_adjust_weights)
        print(edges_to_adjust)
    if num_cycles == 0:
        continue

    c = matrix(np.array(edges_to_adjust_weights))
    G2 = matrix(-1*np.array(cycles, dtype=float))
    h = matrix(-1 * np.ones(num_cycles))
    I = set()
    B = set(range(num_edges))

    print()
    print('------------------------------- Solving ILP -------------------------------')
    (status,x) = ilp(c,G2,h,B=B)
    print('---------------------------------------------------------------------------')
    if DEBUG:
        print(x)
    x = np.array(x)
    adjusted_edge_idx = np.where(x==1)[0]
    num_adjusted[-1] += len(adjusted_edge_idx)
    for e_idx in adjusted_edge_idx:
        node_o, node_i = edges_to_adjust[e_idx]
        g.graph[node_o]['out'].remove(node_i)
        g.graph[node_i]['in'].remove(node_o)

        # count adjusted edges
        for e in eq_edges_map[(node_o, node_i)]:
            n1, n2 = e
            r1 = set(builder.relations[n1])
            r2 = set(builder.relations[n2])
            if not (r1 & r2): continue
            common_pop = (r1 & r2).pop()
            key = tuple(sorted(list(r1 | r2)))
            k_idx = key.index(common_pop)
            value = triangles[key]['D'][k_idx]
            total_edges_adjusted.append((key[(k_idx+1)%3], key[(k_idx+2)%3], key[k_idx], value))

if DEBUG:
    print('num adjusted', num_adjusted)
print()
print('Adjusted values', total_edges_adjusted)
print('Total adjusted:', len(total_edges_adjusted))

if DEBUG:
    print('------------------------')
scc = g.SCC(size=1)
num_scc = len(scc)
eq_scc = eq_scc0 
for scc_idx, s in enumerate(scc):
    # if len(s) <= 2: continue
    if not equal_vertices_judgement(g, s):
        print('This scc are not eq:', scc_idx)
        sys.exit(0)
    if DEBUG:
        print('--------- scc ----------')
    represent = s[-1]
    for idx in s:
        if idx in eq_scc:
            represent = idx
            break
    s.remove(represent)
    for idx in s:
        if DEBUG:
            print(idx, builder.relations[idx])
        for node in g.graph[idx]['in']:
            g.graph[represent]['in'].add(node)
            g.graph[node]['out'].add(represent)
            g.graph[node]['out'].remove(idx)
        for node in g.graph[idx]['out']:
            g.graph[represent]['out'].add(node)
            g.graph[node]['in'].add(represent)
            g.graph[node]['in'].remove(idx)
        del g.graph[idx]
        if idx in g.start_nodes:
            g.start_nodes.discard(idx)
        else:
            g.start_nodes.discard(represent)
    g.graph[represent]['in'].remove(represent)
    g.graph[represent]['out'].remove(represent)
    if represent in eq_scc:
        eq_scc[represent].extend(s)
    else:
        eq_scc[represent] = s
    if DEBUG:
        print(g.graph[represent])
    
#    print(edges_to_adjust)
if DEBUG:
    print('------------------------')
    for node in g.graph:
        if len(g.graph[node]['in']) == 0:
            print(node)
    print(g.start_nodes)

print('eq_scc', eq_scc)
print()
print('Building the tree...')
builder.build_tree(eq_scc=eq_scc, D=None, **args)
print()
print('Plotting...')
builder.export_graph(save_path, print_gene_flow=print_gene_flow)
print()
print('Saved to %s' % save_path)

print()
print('Done.')