from itertools import combinations
from data_structures import UnionFindSet, Graph 
from anytree import Node, PreOrderIter, RenderTree
from anytree.exporter import DotExporter
from collections import defaultdict
# from sklearn.linear_model import LinearRegression
import numpy as np
import sys
from geneflow_P import Probability_distribution

DEBUG = False
class TreeBuilder(object):
    def __init__(self, populations, triangles, s2_record):
        self.triangles = triangles
        self.populations = sorted(populations)
        self.num_populations = len(self.populations)
        self.relations = list(combinations(self.populations, 2))
        self.s2_record = s2_record
        self.root = None 
        self.gene_flow = None 

    def D2graph(self, D):
        g = Graph(len(self.relations))
        for i, rule in enumerate(D):
            node1 = tuple(sorted([rule[0], rule[2]]))
            node2 = tuple(sorted([rule[1], rule[2]]))
            relation = rule[3]
            if relation == 10:
                g.addEdge(self.relations.index(node1), self.relations.index(node2))
            elif relation == -10:
                g.addEdge(self.relations.index(node2), self.relations.index(node1))
            elif -10 < relation and relation < 10:
                g.addEdge(self.relations.index(node1), self.relations.index(node2), equal=True, value=rule[4])
            else:
                raise ValueError('relation must be 1, -1 or 0, but got {}'.format(relation))
        self.g = g
        return g 
    
    def get_order_from_graph_back(self, D):
        # can be speed up by combining with build_tree
        def helper(queue, order):
            if len(queue) == 0: return
            node = queue.pop()
            order.append([node] + list(g.graph[node]['eq']))
            for out_node in g.graph[node]['out']:
                g.graph[out_node]['in'].remove(node)
                if len(g.graph[out_node]['in']) == 0:
                    queue.add(out_node)
            del g.graph[node]
            helper(queue, order)
    
        g = self.D2graph(D)
        queue = g.start_nodes
        order = []
        helper(queue, order)
        if len(g.graph) > 0:
            raise ValueError('Graph is not a DAG')
        return order

    def get_order_from_graph(self, eq_scc=None, D=None):
        # can be speed up by combining with build_tree
        def helper(queue, order):
            if len(queue) == 0: return
            nodes = queue.pop(0)
            order.append(nodes)
            next_nodes = []
            for node in nodes:
                for out_node in g.graph[node]['out']:
                    g.graph[out_node]['in'].remove(node)
                    tmp_eq = set()
                    for node_eq, _ in g.graph[out_node]['eq']:
                        tmp_eq.add(node_eq)
                    if len(g.graph[out_node]['in']) == 0:
                        next_nodes.append(out_node)
                    elif (g.graph[out_node]['in'] | tmp_eq) == tmp_eq:
                        next_nodes.append(out_node)
                        for node_i in g.graph[out_node]['in']:
                            g.graph[node_i]['out'].remove(out_node)
                del g.graph[node]
            if len(next_nodes) > 0:
                queue.append(next_nodes)
            helper(queue, order)
        if self.g is None: 
            self.D2graph(D)

        g = self.g
        queue = [list(g.start_nodes)]
        order = []
        helper(queue, order)
        if len(g.graph) > 0:
            print('graph', g.graph)
            raise ValueError('Graph is not a DAG')
        for o_id, o in enumerate(order):
            for node in o:
                if node in eq_scc: 
                    order[o_id].extend(eq_scc[node])

        if DEBUG:
            print('order', order)
            print([[self.relations[idx] for idx in o] for o in order])
        return order

    def get_order_from_degree(self, D):
        count = [0] * len(self.relations)
        for i, rule in enumerate(D):
            node1 = tuple(sorted([rule[0], rule[2]]))
            node2 = tuple(sorted([rule[1], rule[2]]))
            relation = rule[3]
            if relation == 1:
                count[self.relations.index(node2)] += 1
            elif relation == -1:
                count[self.relations.index(node1)] += 1
            elif relation == 0:
                continue
            else:
                raise ValueError('relation must be 1, -1 or 0, but got {}'.format(relation))
        reverse_index_count = [[] for _ in range(self.num_populations)]
        for i, c in enumerate(count):
            reverse_index_count[c].append(i)
        return reverse_index_count
    
    def record_gene_flow(self, visited_leaves, l1, l2, order_id):
        if visited_leaves[l1]['id'] > visited_leaves[l2]['id']:
            if visited_leaves[l1]['last_gene_flow_id'] < order_id:
                visited_leaves[l1]['last_gene_flow_id'] = order_id
            visited_leaves[l1]['gene_flow'].append(l2)
        elif visited_leaves[l1]['id'] < visited_leaves[l2]['id']:
            if visited_leaves[l2]['last_gene_flow_id'] < order_id:
                visited_leaves[l2]['last_gene_flow_id'] = order_id
            visited_leaves[l2]['gene_flow'].append(l1)
        else:
            if visited_leaves[l1]['last_gene_flow_id'] < order_id:
                visited_leaves[l1]['last_gene_flow_id'] = order_id
            if visited_leaves[l2]['last_gene_flow_id'] < order_id:
                visited_leaves[l2]['last_gene_flow_id'] = order_id
            visited_leaves[l1]['gene_flow'].append(l2)
            # visited_leaves[l2]['gene_flow'].append(l1)

    def add_gene_flow(self, gene_flow, visited_leaves, l1, l2, order_id, gene_flow_propotion_population, threshold=0):
        node_l1 = visited_leaves[l1]['node']
        node_l2 = visited_leaves[l2]['node']
        if node_l1 is None or node_l2 is None:
            return
        # print('gene flow in {} and {}'.format(l1, l2))
        # print(gene_flow)
        # print(node_l1.depth)
        # print(node_l2.depth)
        # print(visited_leaves[l1]['gene_flow'])
#        print('---------- propotion ----------')
        if node_l1.depth > node_l2.depth and visited_leaves[l2]['last_gene_flow_id'] < order_id:
            for leaf in visited_leaves[l2]['gene_flow']:
                if abs(self.get_value(l1, leaf, l2)) > threshold:
                    gene_flow.append((l2, leaf))
#                print(l2, leaf, l1)
                gene_flow_propotion_population.append((l2, leaf, l1))
            visited_leaves[l2]['gene_flow'] = []
        elif node_l1.depth < node_l2.depth and visited_leaves[l1]['last_gene_flow_id'] < order_id:
            for leaf in visited_leaves[l1]['gene_flow']:
                if abs(self.get_value(l2, leaf, l1)) > threshold:
                    gene_flow.append((l1, leaf))
#                print(l1, leaf, l2)
                gene_flow_propotion_population.append((l1, leaf, l2))
            visited_leaves[l1]['gene_flow'] = []
        elif node_l1.depth == node_l2.depth and visited_leaves[l1]['last_gene_flow_id'] < order_id and visited_leaves[l2]['last_gene_flow_id'] < order_id:
            for leaf in visited_leaves[l1]['gene_flow']:
                if abs(self.get_value(l2, leaf, l1)) > threshold:
                    gene_flow.append((l1, leaf))
#                print(l1, leaf, l2)
                gene_flow_propotion_population.append((l1, leaf, l2))
            for leaf in visited_leaves[l2]['gene_flow']:
                if abs(self.get_value(l1, leaf, l2)) > threshold:
                    gene_flow.append((l2, leaf))
#                print(l2, leaf, l1)
                gene_flow_propotion_population.append((l2, leaf, l1))
            visited_leaves[l1]['gene_flow'] = []
            visited_leaves[l2]['gene_flow'] = []

    def get_value(self, A, B, C, typ='D'):
        key = tuple(sorted([A, B, C]))
        pos = key.index(C)
        sign = 1 if (key.index(A) - pos) % 3 < (key.index(B) - pos) % 3 else -1
        value = sign * self.triangles[key][typ][pos]
        return value
    
    def merge_gene_flow(self, gene_flow, visited_leaves, threshold=100):
        def merge(merged_gene_flow, parents, threshold):
            for p, nodes in parents.items():
                merge_flag = True
                for node1, node2 in combinations(nodes, 2):
                    # check D statistics
                    A, B, C = old_l1, node1.name, node2.name
                    key = tuple(sorted([A, B, C]))
                    value = self.triangles[key]['D'][key.index(A)]
                    if abs(value) < threshold:
                        merge_flag = False

                if merge_flag and len(nodes) > 1 and visited_leaves[old_l1]['node'].parent != p:
                    if len(nodes) == len(p.children):
                        new_node = p
                    else:
                        new_node = Node(int(p.name) - 0.5, parent=p)
                        if DEBUG:
                            print('%s --1--> %s'% (p.name, str(p.name - 0.5)))
                    merged_gene_flow.append((old_l1, new_node.name))
                    for n in nodes:
                        n.parent = new_node
                        if DEBUG:
                            print('%s --2--> %s'% (new_node.name, n.name))
                else:
                    for n in nodes:
                        merged_gene_flow.append((old_l1, n.name))

        if DEBUG:
            print('=============================================================')
        merged_gene_flow = []
        gene_flow = sorted(gene_flow)
        if DEBUG:
            print(gene_flow)
        parents = defaultdict(set)
        old_l1 = None
        for i, (l1, l2) in enumerate(gene_flow):
            if DEBUG:
                print(l1, l2)
            if l1 != old_l1:
                merge(merged_gene_flow, parents, threshold)
                parents = defaultdict(set)
                node_l2 = visited_leaves[l2]['node']
                parents[node_l2.parent].add(node_l2)
                if DEBUG:
                    print('parents', parents)
                old_l1 = l1
            else:
                node_l2 = visited_leaves[l2]['node']
                parents[node_l2.parent].add(node_l2)
                if DEBUG:
                    print('parents', parents)
        merge(merged_gene_flow, parents, threshold)
        print()
        print('merged_gene_flow', merged_gene_flow)
        return merged_gene_flow

    def build_tree(self, eq_scc=None, D=None, **kwargs):
        threshold2 = kwargs.get('threshold2', 0)
        gene_flow_threshold = kwargs.get('gene_flow_threshold', 100)

        order = self.get_order_from_graph(eq_scc, D)
        if DEBUG:
            print('order', order)
        visited_leaves = {} # {leaf: {'node': Node class, 'i100d': order_id, 'gene_flow': gene_flow}}
        glue_id = 0
        # pic_id = 0
        cyclic_flag = False 
        gene_flow = [] 
        gene_flow_propotion_population = []
        tmp_gene_flow = []
        for order_id, edges in enumerate(order):
            # if there are abruptions in the order, then the tree is cyclic
            if len(edges) != 0 and cyclic_flag:
                raise ValueError('cyclic graph')
            elif edges == []:
                cyclic_flag = True 

            nodes_to_be_clustered = []
            for e in edges:
                l1, l2 = self.relations[e]
                nodes_to_be_clustered.append([l1, l2])
#                if e == 10:
#                   print('out of add gene flow', l1, l2)
                if l1 in visited_leaves and l2 in visited_leaves:
                    self.add_gene_flow(tmp_gene_flow, visited_leaves, l1, l2, order_id, gene_flow_propotion_population, threshold2)
                if l1 not in visited_leaves:
                    visited_leaves.update({l1: {'node': None, 'id': order_id, 'gene_flow': [], 'last_gene_flow_id': 0}})
                if l2 not in visited_leaves:
                    visited_leaves.update({l2: {'node': None, 'id': order_id, 'gene_flow': [], 'last_gene_flow_id': 0}})
                # gene flow record
                self.record_gene_flow(visited_leaves, l1, l2, order_id)
            # cluster the nodes with unionfindset, the clustered nodes will have the same parent (i.e. n-ary tree)
            ufs = UnionFindSet(nodes_to_be_clustered)
            # print(nodes_to_be_clustered)
            ufs.init_union()

            for cluster in ufs.different_sets.values():
                # print('cluster', cluster)
                glue_node = Node(glue_id)
                glue_id += 1
                for leaf in cluster:
                    if visited_leaves[leaf]['node'] is None:
                        new_node = Node(leaf, parent=glue_node)
                        visited_leaves[leaf]['node'] = new_node
                    else:
                        tmp_node = visited_leaves[leaf]['node']
                        if tmp_node.root != glue_node:
                            tmp_node.root.parent = glue_node
                if len(glue_node.children) <= 1:
                    glue_node.children[0].parent = None
                    glue_id -= 1

            # root = visited_leaves[populations[0]]['node'].root
            # RenderTreeGraph(root).to_picture("tree_%d.png" % pic_id)
            # pic_id += 1
        print()
        print('all gene flows', tmp_gene_flow)
        if len(tmp_gene_flow) > 0:
            if DEBUG:
                print('tmp_gene_flow', tmp_gene_flow)
            tmp_gene_flow = self.calculate_propotion(tmp_gene_flow, order, visited_leaves, **kwargs)
            merged_gene_flow = self.merge_gene_flow(tmp_gene_flow, visited_leaves, gene_flow_threshold)
            gene_flow = merged_gene_flow
        root = visited_leaves[self.populations[0]]['node'].root
        self.root = root
        self.gene_flow = gene_flow
        if DEBUG:
            print(gene_flow)
        # self.gene_flow = self.merge_gene_flow(gene_flow, self.root)
        # print('gene_flow:\n', self.gene_flow)
        # self.calculate_propotion(gene_flow_propotion_population)
        return root
    
    def calculate_propotion(self, gene_flow, order, visited_leaves, **kwargs):
        def find_s1_(s1, tar, s2):
            def find_first_common_ancestor(s1, tar):
                s1_node = visited_leaves[s1]['node']
                tar_node = visited_leaves[tar]['node']
                s1_ancestors = []
                tar_ancestors = []
                while s1_node.parent is not None:
                    s1_ancestors.append(s1_node.parent)
                    s1_node = s1_node.parent
                while tar_node.parent is not None:
                    tar_ancestors.append(tar_node.parent)
                    tar_node = tar_node.parent
                common_ancestor = None
                for node_idx, s1_ancestor in enumerate(s1_ancestors):
                    if s1_ancestor in tar_ancestors:
                        common_ancestor = s1_ancestor
                        break
                s1__block_nodes = [tar_ancestors[tar_ancestors.index(common_ancestor) - 1]]
                # s1__block_nodes = tar_ancestors[tar_ancestors.index(common_ancestor)-1:tar_ancestors.index(common_ancestor)]
                return common_ancestor, s1__block_nodes
            def bfs_s1_(queue, block_nodes, visited):
                while queue:
                    if DEBUG:
                        print('--------------------')
                        print(queue)
                    node = queue.pop(0)
                    visited.append(node)
                    if node in block_nodes:
                        continue
                    #if node.is_leaf and \
                    #    (((node.name, s1) not in gene_flow and (s1, node.name) not in gene_flow) or \
                    #    ((node.name, tar) not in gene_flow and (tar, node.name) not in gene_flow)):
                    if node.is_leaf:
                        return node.name
                    if node.parent and node.parent not in visited:
                        queue.append(node.parent)
                    for child in node.children:
                        if child not in visited:
                            queue.append(child)
                return None
            s1_node = visited_leaves[s1]['node']
            tar_node = visited_leaves[tar]['node']
            s2_node = visited_leaves[s2]['node']
            common_ancestor, s1__block_nodes = find_first_common_ancestor(s1, tar)
            s1__block_nodes.extend([s1_node, tar_node, s2_node])
            if DEBUG:
                print(common_ancestor)
                print(s1__block_nodes)
            queue = [tar_node.parent]
            queue = [s1_node.parent]
            visited = []
            s1_ = bfs_s1_(queue, s1__block_nodes, visited)
            return s1_

        def calculate_m1(s1, tar, s2, pd=False):
            if DEBUG:
                print(s1, tar)
            s1_ = find_s1_(s1, tar, s2)
            if DEBUG:
                print('s1: %s, tar: %s, s2: %s, s1_: %s' % (s1, tar, s2, s1_))
            if s2 is None or s1_ is None:
                return None, None
            key = tuple(sorted((s1, s2, tar)))
            pos = key.index(s1)
            sign = 1 if (key.index(s2) - pos) % 3 < (key.index(tar) - pos) % 3 else -1
            D1 = sign * self.triangles[key]['f4'][pos]
            Z1 = sign * self.triangles[key]['D'][pos]


            key = tuple(sorted((s1, s2, s1_)))
            pos = key.index(s1)
            sign = 1 if (key.index(s2) - pos) % 3 < (key.index(s1_) - pos) % 3 else -1
            D2 = sign * self.triangles[key]['f4'][pos]
            Z2 = sign * self.triangles[key]['D'][pos]
            # m1 = D1 / D2

            if pd:
                mix, m_pd = Probability_distribution(D1,D2,Z1,Z2)
            else:
                mix = D1 / D2
                m_pd = None

            if DEBUG:
                print('D1: %f, D2: %f' % (D1, D2))
            return mix, m_pd

        bidirectional = kwargs.get('bidirectional', True)
        calculate_pd = kwargs.get('calculate_pd', False)
        tar_block_list = kwargs.get('target_block_list', [])
        s1_block_list = kwargs.get('s1_block_list', [])
        # calculate_m1('Dushan', 'Baojianshan')
        # calculate_m1('Baojianshan', 'Dushan')
        # calculate_m1('Boshan', 'Dushan')
        # print(self.s1_record)
        propotion = {}
        print('Calculating gene flows...')
        gf_cnt = 0
        for s1, tar in gene_flow:
            if tar in tar_block_list: continue
            if s1 in s1_block_list: continue
            propotion[(s1, tar)] = []
            if (s1, tar) not in self.s2_record:
                continue
            for s2 in self.s2_record[(s1, tar)]:
                if DEBUG:
                    print('-----------------------------------------------') 
                # pd = False to make it fast
                mix_bc, m_pd_bc = calculate_m1(s1, tar, s2)
                if DEBUG:
                    print('mix:', mix_bc)
                # mix_a1b, mix_pd_a1b = calculate_m1(s2, tar, s1)
                # print('mix:', mix_a1b)
                # if mix_bc is None or mix_a1b is None: continue
                if mix_bc is None: continue
                # if mix_bc > 1 or mix_bc < 0 or mix_a1b > 1 or mix_a1b < 0: continue
                if mix_bc > 1 or mix_bc < 0: continue
                print('Calculating gene flow %d' % gf_cnt)
                if calculate_pd:
                    mix_bc, m_pd_bc = calculate_m1(s1, tar, s2, pd=True)
                else:
                    m_pd_bc = float('nan')
                propotion[(s1, tar)].append((s2, mix_bc, m_pd_bc))
                gf_cnt += 1

        print()
        gf_cnt = 0
        if bidirectional:
            print('Calculating reversed gene flows...')
            for tar, s1 in gene_flow:
                if s1 in s1_block_list: continue
                if tar in tar_block_list: continue
                propotion[(s1, tar)] = []
                if (s1, tar) not in self.s2_record:
                    continue
                for s2 in self.s2_record[(s1, tar)]:
                    mix_bc, m_pd_bc = calculate_m1(s1, tar, s2)
                    # mix_a1b, mix_pd_a1b = calculate_m1(s2, tar, s1)
                    # print('mix_bc, mix_a1b', mix_bc, mix_a1b)
                    # if mix_bc is None or mix_a1b is None: continue
                    if mix_bc is None: continue
                    if mix_bc > 1 or mix_bc < 0: continue
                    # if mix_bc > 1 or mix_bc < 0 or mix_a1b > 1 or mix_a1b < 0: continue
                    print('Calculating gene flow %d' % gf_cnt)
                    if calculate_pd:
                        mix_bc, m_pd_bc = calculate_m1(s1, tar, s2, pd=True)
                    else:
                        m_pd_bc = float('nan')
                    propotion[(s1, tar)].append((s2, mix_bc, m_pd_bc))
                    gf_cnt += 1
        # for s2, tar in gene_flow:
        #     if tar in tar_block_list: continue
        #     propotion[(s2, tar)] = []
        #     if (s2, tar) not in self.s1_record:
        #         continue
        #     for s1 in self.s1_record[(s2, tar)]:
        #         if s1 in s1_block_list: continue
        #         if DEBUG:
        #             print('-----------------------------------------------') 
        #         # pd = False to make it fast
        #         mix_bc, m_pd_bc = calculate_m1(s1, tar, s2)
        #         if DEBUG:
        #             print('mix:', mix_bc)
        #         # mix_a1b, mix_pd_a1b = calculate_m1(s2, tar, s1)
        #         # print('mix:', mix_a1b)
        #         # if mix_bc is None or mix_a1b is None: continue
        #         if mix_bc is None: continue
        #         # if mix_bc > 1 or mix_bc < 0 or mix_a1b > 1 or mix_a1b < 0: continue
        #         if mix_bc > 1 or mix_bc < 0: continue
        #         print('Calculating gene flow %d' % gf_cnt)
        #         if calculate_pd:
        #             mix_bc, m_pd_bc = calculate_m1(s1, tar, s2, pd=True)
        #         else:
        #             m_pd_bc = float('nan')
        #         propotion[(s2, tar)].append((s1, mix_bc, m_pd_bc))
        #         gf_cnt += 1

        # print()
        # gf_cnt = 0
        # if bidirectional:
        #     print('Calculating reversed gene flows...')
        #     for tar, s2 in gene_flow:
        #         if tar in tar_block_list: continue
        #         propotion[(s2, tar)] = []
        #         if (s2, tar) not in self.s1_record:
        #             continue
        #         for s1 in self.s1_record[(s2, tar)]:
        #             if s1 in s1_block_list: continue
        #             mix_bc, m_pd_bc = calculate_m1(s1, tar, s2)
        #             # mix_a1b, mix_pd_a1b = calculate_m1(s2, tar, s1)
        #             # print('mix_bc, mix_a1b', mix_bc, mix_a1b)
        #             # if mix_bc is None or mix_a1b is None: continue
        #             if mix_bc is None: continue
        #             if mix_bc > 1 or mix_bc < 0: continue
        #             # if mix_bc > 1 or mix_bc < 0 or mix_a1b > 1 or mix_a1b < 0: continue
        #             print('Calculating gene flow %d' % gf_cnt)
        #             if calculate_pd:
        #                 mix_bc, m_pd_bc = calculate_m1(s1, tar, s2, pd=True)
        #             else:
        #                 m_pd_bc = float('nan')
        #             propotion[(s2, tar)].append((s1, mix_bc, m_pd_bc))
        #             gf_cnt += 1

        print()
        print('---------------------------------------------------------------') 
        # mat = "{:30}\t{:1}\t{:1}"
        mat = "{:10}\t{:10}\t{:10}\t{:1}\t{:1}"
        print(mat.format('s2', 'tar', 's1', 'mix', 'm_pd'))
        mat = "{:10}\t{:10}\t{:10}\t{:.3f}\t{:.3f}"
        

        m_pd_display_threshold = kwargs.get('gene_flow_display_confidence', 0.0)
        gene_flows_checked_by_propotion = []
        for gene_flow, m1s in propotion.items():
            s1, tar = gene_flow
            for m1 in m1s:
                s2, mix, m_pd = m1
                # print(s2, tar, s1, mix, m_pd)
                print(mat.format(s2, tar, s1, mix, m_pd))
                if (s1, tar) not in gene_flows_checked_by_propotion:
                    if not (calculate_pd and m_pd < m_pd_display_threshold):
                        gene_flows_checked_by_propotion.append((s1, tar))
        # print(',,,', gene_flows_checked_by_propotion)
        # print(m_pd_display_threshold)
        return gene_flows_checked_by_propotion

    def calculate_propotion_bak2(self, gene_flow, order, visited_leaves, bidirectional=True):
        def find_s2_s1_(s1, tar):
            def find_first_common_ancestor(s1, tar):
                s1_node = visited_leaves[s1]['node']
                tar_node = visited_leaves[tar]['node']
                s1_ancestors = []
                tar_ancestors = []
                while s1_node.parent is not None:
                    s1_ancestors.append(s1_node.parent)
                    s1_node = s1_node.parent
                while tar_node.parent is not None:
                    tar_ancestors.append(tar_node.parent)
                    tar_node = tar_node.parent
                common_ancestor = None
                for node_idx, s1_ancestor in enumerate(s1_ancestors):
                    if s1_ancestor in tar_ancestors:
                        common_ancestor = s1_ancestor
                        break
                s2_block_nodes = s1_ancestors[node_idx-1:node_idx]
                s1__block_nodes = [common_ancestor]
                # s1__block_nodes = tar_ancestors[tar_ancestors.index(common_ancestor)-1:tar_ancestors.index(common_ancestor)]
                return common_ancestor, s2_block_nodes, s1__block_nodes
            def bfs_s1_(queue, block_nodes):
                while queue:
                    if DEBUG:
                        print('--------------------')
                        print(queue)
                    node = queue.pop(0)
                    if node in block_nodes:
                        continue
                    #if node.is_leaf and \
                    #    (((node.name, s1) not in gene_flow and (s1, node.name) not in gene_flow) or \
                    #    ((node.name, tar) not in gene_flow and (tar, node.name) not in gene_flow)):
                    if node.is_leaf:
                        return node.name
                    if node.parent:
                        queue.append(node.parent)
                    for child in node.children:
                        queue.append(child)
                return None
            def bfs_s2(queue, block_nodes):
                while queue:
                    if DEBUG:
                        print('--------------------')
                        print(queue)
                    node = queue.pop(0)
                    if node in block_nodes:
                        continue
                    if node.is_leaf:
                        return node.name
                    if node.parent:
                        queue.append(node.parent)
                    for child in node.children:
                        queue.append(child)
                return None 
            s1_node = visited_leaves[s1]['node']
            tar_node = visited_leaves[tar]['node']
            common_ancestor, s2_block_nodes, s1__block_nodes = find_first_common_ancestor(s1, tar)
            s2_block_nodes.extend([s1_node, tar_node])
            s1__block_nodes.extend([s1_node, tar_node])
            if DEBUG:
                print(common_ancestor)
                print(s2_block_nodes)
                print(s1__block_nodes)
            queue = [tar_node.parent]
            s2 = bfs_s2(queue, s2_block_nodes)
            queue = [s1_node.parent]
            s1_ = bfs_s1_(queue, s1__block_nodes)
            return s2, s1_

        def calculate_m1(s1, tar):
            print('-----------------------------------------------') 
            if DEBUG:
                print(s1, tar)
            s2, s1_ = find_s2_s1_(s1, tar)
            print('s1: %s, tar: %s, s2: %s, s1_: %s' % (s1, tar, s2, s1_))
            if s2 is None or s1_ is None:
                return float('nan')
            key = tuple(sorted((s1, s2, tar)))
            pos = key.index(s1)
            sign = 1 if (key.index(s2) - pos) % 3 < (key.index(tar) - pos) % 3 else -1
            D1 = sign * self.triangles[key]['f4'][pos]

            key = tuple(sorted((s1, s2, s1_)))
            pos = key.index(s1)
            sign = 1 if (key.index(s2) - pos) % 3 < (key.index(s1_) - pos) % 3 else -1
            D2 = sign * self.triangles[key]['f4'][pos]
            m1 = D1 / D2

            print('D1: %f, D2: %f' % (D1, D2))
            return m1

        # calculate_m1('Dushan', 'Baojianshan')
        # calculate_m1('Baojianshan', 'Dushan')
        # calculate_m1('Boshan', 'Dushan')
        propotion = {}
        for s1, tar in gene_flow:
            # s2 == s1 is impossible, otherwise there mustn't be a gene flow from s1 to tar
            m1 = calculate_m1(s1, tar)
            propotion[(s1, tar)] = {'forward': m1, 'backward': float('nan')}

        if bidirectional:
            for tar, s1 in gene_flow:
                m1 = calculate_m1(s1, tar)
                propotion[(tar, s1)]['backward'] = m1
        print('-----------------------------------------------') 
        mat = "{:30}\t{:1}\t{:1}"
        print(mat.format('gene flow', 'forward', 'backward'))
        mat = "{:30}\t{:.3f}\t{:.3f}"
        
        gene_flows_checked_by_propotion = []
        for gene_flow, m1 in propotion.items():
            print(mat.format(str(gene_flow), m1['forward'], m1['backward']))
            if m1['forward'] and m1['forward'] > 0 and m1['forward'] < 1:
                gene_flows_checked_by_propotion.append(gene_flow)
            if m1['backward'] and m1['backward'] > 0 and m1['backward'] < 1:
                gene_flows_checked_by_propotion.append((gene_flow[1], gene_flow[0]))
        return gene_flows_checked_by_propotion


    def calculate_propotion_bak(self, gene_flow, order, bidirectional=True):
        print(gene_flow)
        def find_nearest_population(tar, block_nodes, order):
            for nodes in order:
                for n in nodes:
                    if tar in self.relations[n]:
                        is_block_node = False
                        for block_node in block_nodes:
                            if block_node in self.relations[n]:
                                is_block_node = True
                                break
                        if is_block_node: continue
                        return self.relations[n][0] if self.relations[n][0] != tar else self.relations[n][1]

        def calculate_m1(s1, tar):
            print('------------------------------')
            print(s1, tar)
            s2 = find_nearest_population(tar, [s1], order)
            key = tuple(sorted((s1, s2, tar)))
            pos = key.index(s1)
            sign = 1 if (key.index(s2) - pos) % 3 < (key.index(tar) - pos) % 3 else -1
            D1 = sign * self.triangles[key]['f4'][pos]

            s1_ = find_nearest_population(s1, [tar, s2], order)
            key = tuple(sorted((s1, s2, s1_)))
            pos = key.index(s1)
            sign = 1 if (key.index(s2) - pos) % 3 < (key.index(s1_) - pos) % 3 else -1
            D2 = sign * self.triangles[key]['f4'][pos]
            m1 = D1 / D2

            print('s1: %s, tar: %s, s2: %s, s1_: %s' % (s1, tar, s2, s1_))
            print('D1: %f, D2: %f' % (D1, D2))
            return m1

        propotion = {}
        for s1, tar in gene_flow:
            # s2 == s1 is impossible, otherwise there mustn't be a gene flow from s1 to tar
            m1 = calculate_m1(s1, tar)
            propotion[(s1, tar)] = {'forward': m1, 'backward': None}

        if bidirectional:
            for tar, s1 in gene_flow:
                m1 = calculate_m1(s1, tar)
                propotion[(tar, s1)]['backward'] = m1
        print('---------------------') 
        mat = "{:30}\t{:1}\t{:1}"
        print(mat.format('gene flow', 'forward', 'backward'))
        mat = "{:30}\t{:.3f}\t{:.3f}"
        for gene_flow, m1 in propotion.items():
            print(mat.format(str(gene_flow), m1['forward'], m1['backward']))


#    def calculate_propotion_regression(self, gene_flow_propotion_population):
#        print('calculate propotion')
#        out_group = set()
#        all_statistics = dict()
#        with open('data/D_ai.log') as f:
#            for l in f.readlines():
#                if not l.startswith('result'): continue
#                A, B, C, D = l.split()[1:5]
#                for population in [A, B, C, D]: 
#                    if population not in self.populations:
#                        out_group.add(population)
#                key = tuple(sorted((A, B, C, D)))
#                f4 = float(l.split()[5])
#                all_statistics[key] = f4 
#        out_groups = list(combinations(out_group, 2))
#        for test, R1, R2 in gene_flow_propotion_population:
#            print(test, R1, R2)
#            x = []
#            y = []
#            for O1, O2 in out_groups:
#                # print(test, R1, O1, O2)
#                y.append(all_statistics[tuple(sorted((test, R1, O1, O2)))])
#                x.append(all_statistics[tuple(sorted((test, R2, O1, O2)))])
#            x = np.array(x).reshape(-1, 1)
#            y = np.array(y).reshape(-1, 1)
#            model = LinearRegression()
#            model.fit(x, y)
#
#            k = model.coef_[0][0]
#            b = model.intercept_[0]
#            R2 = model.score(x, y)
#
#            print("k", k)
#            print("a", 1 / (k))
#            print("b", b)
#            print("R2", R2)

    def get_dotexporter(self, print_gene_flow=True):
        options = [] 
        if print_gene_flow:
            for l1, l2 in self.gene_flow:
                options.append("%s -> %s [color=red]" % (l1, l2))
        # print(self.populations)
        # print(self.relations)
        # print(RenderTree(self.root))
        return DotExporter(self.root, options=options)
    
    def export_dot(self, filename, print_gene_flow=True):
        self.get_dotexporter(print_gene_flow).to_dotfile(filename)

    def export_graph(self, filename, print_gene_flow=True):
        self.get_dotexporter(print_gene_flow).to_picture(filename)
