from preprocess import get_adjusted_D_statistics
from itertools import combinations
import copy

def read_data(data_path, threshold, populations):
    D_statistics = []
    adjusted_D_values = []
    triangles = {}
    redstar_indices = {}
    key_visited = []
    eq_scc = None
    min_redstar = []
    s2_record = {}
    with open(data_path) as f:
        for l in f.readlines():
            if not l.startswith('result'): continue
            A, B, C = l.split()[1:4]
            if A not in populations or B not in populations or C not in populations: continue
            f4 = float(l.split()[5])
            value = float(l.split()[6])
            get_adjusted_D_statistics(D_statistics, adjusted_D_values, triangles, A, B, C, f4, value, threshold, min_redstar, redstar_indices, key_visited, s2_record)
            if len(D_statistics) >= len(populations) * (len(populations) - 1) * (len(populations) - 2) / 2:
                break
    return D_statistics, adjusted_D_values, triangles, redstar_indices, min_redstar, s2_record

def equal_vertices_judgement(g, vertices):
    edges = list(combinations(vertices, 2))
    for v1, v2 in edges:
#        print(v1, v2)
#        print(g.graph[v1]['eq'])
#        print(g.graph[v2]['eq'])
        if v2 in [eq[0] for eq in g.graph[v1]['eq']] or v1 in [eq[0] for eq in g.graph[v2]['eq']]: continue
        if v2 in g.graph[v1]['out'] or v2 in g.graph[v1]['in']: return False
    return True


from collections import defaultdict

class Graph:
    def __init__(self, vertices):
        self.V = vertices
        self.adj = defaultdict(list)

    def addEdge(self, u, v):
        self.adj[u].append(v)

    def DFS(self, v, visited, stack):
        visited[v] = True
        stack.append(v)

        for neighbor in self.adj[v]:
            if not visited[neighbor]:
                if self.DFS(neighbor, visited, stack):
                    return True
            elif neighbor in stack:
                cycle = []
                index = stack.index(neighbor)
                for i in range(index, len(stack)):
                    cycle.append(stack[i])
                return cycle

        stack.pop()
        return False

    def findCycles(self):
        visited = [False] * self.V
        stack = []
        for i in range(self.V):
            if not visited[i]:
                result = self.DFS(i, visited, stack)
                if result != False:
                    return result

import networkx as nx 

def simple_cycles(G, eq_edges=[], edges_to_adjust=[]):
    def _unblock(thisnode, blocked, B):
        stack = {thisnode}
        while stack:
            node = stack.pop()
            if node in blocked:
                blocked.remove(node)
                stack.update(B[node])
                B[node].clear()

    # Johnson's algorithm requires some ordering of the nodes.
    # We assign the arbitrary ordering given by the strongly connected comps
    # There is no need to track the ordering as each node removed as processed.
    # Also we save the actual graph so we can mutate it. We only take the
    # edges because we do not want to copy edge and node attributes here.
    subG = type(G)(G.edges())
    sccs = [scc for scc in nx.strongly_connected_components(subG) if len(scc) > 2]

    # Johnson's algorithm exclude self cycle edges like (v, v)
    # To be backward compatible, we record those cycles in advance
    # and then remove from subG
    for v in subG:
        if subG.has_edge(v, v):
            # yield [v]
            subG.remove_edge(v, v)
    
    cycle_ids = set()

    visited_startnodes = set()
    cycles = []
    while sccs:
        scc = sccs.pop()
        sccG = subG.subgraph(scc)
        # order of scc determines ordering of nodes
        # print('scc', scc)
        for startnode in scc:
            if startnode in visited_startnodes:
                continue
            visited_startnodes.add(startnode)
            # startnode = scc.pop()

            # Processing node runs "circuit" routine from recursive version
            path = [startnode]
            blocked = set()  # vertex: blocked from search?
            closed = set()  # nodes involved in a cycle
            blocked.add(startnode)
            B = defaultdict(set)  # graph portions that yield no elementary circuit

            startnbrs = list(sccG[startnode])
            for node in copy.deepcopy(startnbrs):
                if (startnode, node) in eq_edges:
                    startnbrs.remove(node)
            if not startnbrs:
                continue
            
            # print(startnode, startnbrs)
            stack = [(startnode, copy.deepcopy(startnbrs))]  # sccG gives comp nbrs

            while stack:
                thisnode, nbrs = stack[-1]
                if nbrs:
                    nextnode = nbrs.pop()
                    #print('nextnode', nextnode)
                    if nextnode == startnode:
                        # paths.add(''.join(path[:]))
                        # print(path)
                        edge_id = [0] * len(edges_to_adjust)
                        for node_idx in range(len(path)):
                            edge = (path[node_idx], path[(node_idx + 1) % len(path)])
                            if edge in edges_to_adjust:
                                edge_id[edges_to_adjust.index(edge)] = 1
                        cycle_ids.add(tuple(edge_id))
                        #yield path[:]
                        closed.update(path)
                        # print('path', path)
                        #print('closed', closed)
                    #                        #print "Found a cycle", path, closed
                    elif nextnode not in blocked:
                        path.append(nextnode)
                        stack.append((nextnode, list(sccG[nextnode])))
                        closed.discard(nextnode)
                        blocked.add(nextnode)
                        #print('path', path)
                        continue
                # done with nextnode... look for more neighbors
                if not nbrs:  # no more nbrs
                    if thisnode in closed:
                        _unblock(thisnode, blocked, B)
                    else:
                        for nbr in sccG[thisnode]:
                            if thisnode not in B[nbr]:
                                B[nbr].add(thisnode)
                    stack.pop()
                    #                assert path[-1] == thisnode
                    path.pop()
            # done processing this node
            # print('paths', paths)
            H = subG.subgraph(scc)  # make smaller to avoid work in SCC routine
            H = type(G)(H.edges())
            # print('remove', startnode, startnbrs)
            # for node in startnbrs:
            for node in startnbrs:
                # print('remove', startnode, node)
                H.remove_edge(startnode, node) 
                subG.remove_edge(startnode, node) 
            # for new_scc in nx.strongly_connected_components(H):
                # print('new_scc', new_scc)
            sccs.extend(new_scc for new_scc in nx.strongly_connected_components(H) if len(new_scc) > 2)
            del H
            break
            #print(startnode, 'finished')
    
    return cycle_ids 


if __name__ == '__main__':
#    g = Graph(4)
#    g.addEdge(0, 1)
#    g.addEdge(0, 2)
#    g.addEdge(1, 2)
#    g.addEdge(2, 0)
#    g.addEdge(2, 3)
#    g.addEdge(3, 3)
#
#    print(g.findCycles())

    import networkx as nx

    # Create Directed Graph
    G=nx.DiGraph()

    # Add a list of nodes:
    G.add_nodes_from(["a","b","c","d","e"])

    # Add a list of edges:
    # G.add_edges_from([("a","b"),("b","c"), ("c","a"), ("b","d"), ("d","e"), ("e","a")])
    G.add_edges_from([('b','a'), ('c','b'), ("a","b"),("b","c"), ("c","a"), ("b","d"), ("d","a")])
    # G.add_edges_from([('a','d'), ('a','c'), ('b','a'), ('c','b'), ("a","b"),("b","c"), ("c","a")])

    #Return a list of cycles described as a list o nodes
    print(list(simple_cycles(G, [('a', 'b'), ('b', 'c'), ('b', 'a'), ('c', 'b')], [('c', 'a')])))

#    G=nx.DiGraph()
#
#    # Add a list of nodes:
#    G.add_nodes_from([1,2,3,4,5])
#
#    # Add a list of edges:
#    G.add_edges_from([(1,2), (2,1)])

    #Return a list of cycles described as a list o nodes
#    cycles = list(nx.simple_cycles(G))
#    print(cycles)
#    new_cycles = []
#    for cycle in cycles:
#        edges = []
#        for i in range(len(cycle)):
#            edges.append((cycle[i], cycle[(i+1) % len(cycle)]))
#        new_cycles.append(edges)
#    cycles = new_cycles
#
#    import cvxopt
#    from cvxopt import matrix
#    from cvxopt.glpk import ilp
#    import numpy as np
#
#    num_edges = 200
#    num_cycles = 100
#    edges = list(G.edges)
#    c = matrix(np.ones(num_edges, dtype=float))
#    G2 = matrix(np.zeros((num_cycles, num_edges), dtype=float))
#    print(cycles)
#    print(edges[0])
#    for i in range(num_cycles):
#        for j in range(num_edges):
#            if edges[j % 6] in cycles[i % 2]:
#                G2[i,j] = -1
#    h = matrix(-1 * np.ones(num_cycles, dtype=float))
#    I = set()
#    B = set(range(num_edges))
#    (status,x)=ilp(c,G2,h,B=B)
#    print(status)
#    print(x)
#